from zope.interface import implements

from plone.z3cform.fieldsets.extensible import ExtensibleForm

from plone.autoform.interfaces import IAutoExtensibleForm
from plone.autoform.base import AutoFields

_marker = object()

class AutoExtensibleForm(AutoFields, ExtensibleForm):
    """Mixin class for z3c.form forms that support fields extracted from
    a schema
    """

    implements(IAutoExtensibleForm)
    
    @property
    def schema(self):
        raise NotImplemented("The class deriving from AutoExtensibleForm must have a 'schema' property")

    @property
    def additionalSchemata(self):
        """Default to there being no additional schemata
        """
        return ()
    
    def updateFields(self):
        self.updateFieldsFromSchemata()
        super(AutoExtensibleForm, self).updateFields()


import unittest
import doctest

import zope.app.testing.placelesssetup
import zope.testing.doctest

def test_suite():
    return unittest.TestSuite((
        zope.testing.doctest.DocFileSuite('../autoform.txt',
            setUp=zope.app.testing.placelesssetup.setUp,
            tearDown=zope.app.testing.placelesssetup.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,),
        zope.testing.doctest.DocFileSuite('../view.txt',
            setUp=zope.app.testing.placelesssetup.setUp,
            tearDown=zope.app.testing.placelesssetup.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,),
        zope.testing.doctest.DocFileSuite('../supermodel.txt',
            setUp=zope.app.testing.placelesssetup.setUp,
            tearDown=zope.app.testing.placelesssetup.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,),
        ))


from zope.interface import implements

from z3c.form.form import DisplayForm
from z3c.form.interfaces import IFormLayer

try:
    from Products.Five.bbb import AcquisitionBBB as Explicit
except ImportError:
    from Acquisition import Explicit

from plone.z3cform import z2

from plone.autoform.interfaces import IWidgetsView
from plone.autoform.base import AutoFields

class WidgetsView(AutoFields, DisplayForm, Explicit):
    """Mix-in to allow widgets (in view mode) to be accesed from browser
    views.
    """
    
    implements(IWidgetsView)
    
    # You should set one or more of these, or the 'fields' variable
    
    schema = None
    additionalSchemata = ()
    
    request_layer = IFormLayer

    def update(self):
        pass

    def render(self):
        if getattr(self, 'index', None) is not None:
            return self.index()
        raise NotImplemented("You must implement the 'render' method")
    
    # Helper methods
    
    def __call__(self):
        z2.switch_on(self)
        self._update()
        self.update()
        return self.render()
    
    def _update(self):
        self.updateFieldsFromSchemata()
        self.updateWidgets()
        
        # shortcut 'widget' dictionary for all fieldsets
        self.w = {}
        for k, v in self.widgets.items():
            self.w[k] = v
        
        groups = []
        self.fieldsets = {}
        
        for idx, groupFactory in enumerate(self.groups):
            group = groupFactory(self.context, self.request, self)
            group.update()
            
            for k, v in group.widgets.items():
                self.w[k] = v
            
            groups.append(group)
            
            group_name = getattr(group, '__name__', str(idx))
            self.fieldsets[group_name] = group
    
        self.groups = tuple(groups)


import unittest

from zope.interface import Interface
import zope.schema

from plone.supermodel.utils import ns

from elementtree import ElementTree

from plone.autoform.interfaces import OMITTED_KEY, WIDGETS_KEY, MODES_KEY, ORDER_KEY
from plone.autoform.interfaces import READ_PERMISSIONS_KEY, WRITE_PERMISSIONS_KEY
from plone.autoform.supermodel import FormSchema, SecuritySchema

class TestFormSchema(unittest.TestCase):
    
    namespace = 'http://namespaces.plone.org/supermodel/form'
    
    def test_read(self):
        field_node = ElementTree.Element('field')
        field_node.set(ns("widget", self.namespace), "SomeWidget")
        field_node.set(ns("mode", self.namespace), "hidden")
        field_node.set(ns("omitted", self.namespace), "true")
        field_node.set(ns("before", self.namespace), "somefield")
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")
        
        handler = FormSchema()
        handler.read(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals({'dummy': 'SomeWidget'}, IDummy.getTaggedValue(WIDGETS_KEY))
        self.assertEquals({'dummy': 'true'}, IDummy.getTaggedValue(OMITTED_KEY))
        self.assertEquals({'dummy': 'hidden'}, IDummy.getTaggedValue(MODES_KEY))
        self.assertEquals([('dummy', 'before', 'somefield',)], IDummy.getTaggedValue(ORDER_KEY))

    def test_read_multiple(self):
        field_node1 = ElementTree.Element('field')
        field_node1.set(ns("widget", self.namespace), "SomeWidget")
        field_node1.set(ns("mode", self.namespace), "hidden")
        field_node1.set(ns("omitted", self.namespace), "true")
        field_node1.set(ns("before", self.namespace), "somefield")
        
        field_node2 = ElementTree.Element('field')
        field_node2.set(ns("mode", self.namespace), "display")
        field_node2.set(ns("omitted", self.namespace), "yes")
        
        class IDummy(Interface):
            dummy1 = zope.schema.TextLine(title=u"dummy1")
            dummy2 = zope.schema.TextLine(title=u"dummy2")
        
        handler = FormSchema()
        handler.read(field_node1, IDummy, IDummy['dummy1'])
        handler.read(field_node2, IDummy, IDummy['dummy2'])
    
        self.assertEquals({'dummy1': 'SomeWidget'}, IDummy.getTaggedValue(WIDGETS_KEY))
        self.assertEquals({'dummy1': 'true', 'dummy2': 'yes'}, IDummy.getTaggedValue(OMITTED_KEY))
        self.assertEquals({'dummy1': 'hidden', 'dummy2': 'display'}, IDummy.getTaggedValue(MODES_KEY))
        self.assertEquals([('dummy1', 'before', 'somefield',)], IDummy.getTaggedValue(ORDER_KEY))
    
    def test_read_no_data(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy1")

        handler = FormSchema()
        handler.read(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals(None, IDummy.queryTaggedValue(WIDGETS_KEY))
        self.assertEquals(None, IDummy.queryTaggedValue(OMITTED_KEY))
        self.assertEquals(None, IDummy.queryTaggedValue(MODES_KEY))
        self.assertEquals(None, IDummy.queryTaggedValue(ORDER_KEY))
        
    def test_write(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy1")
            
        IDummy.setTaggedValue(WIDGETS_KEY, {'dummy': 'SomeWidget'})
        IDummy.setTaggedValue(OMITTED_KEY, {'dummy': 'true'})
        IDummy.setTaggedValue(MODES_KEY, {'dummy': 'hidden'})
        IDummy.setTaggedValue(ORDER_KEY, [('dummy', 'before', 'somefield',)])
        
        handler = FormSchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals("SomeWidget", field_node.get(ns("widget", self.namespace)))
        self.assertEquals("true", field_node.get(ns("omitted", self.namespace)))
        self.assertEquals("hidden", field_node.get(ns("mode", self.namespace)))
        self.assertEquals("somefield", field_node.get(ns("before", self.namespace)))
    
    def test_write_partial(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy1")
        
        IDummy.setTaggedValue(WIDGETS_KEY, {'dummy': 'SomeWidget'})
        IDummy.setTaggedValue(OMITTED_KEY, {'dummy2': 'true'})
        IDummy.setTaggedValue(MODES_KEY, {'dummy': 'display', 'dummy2': 'hidden'})
        IDummy.setTaggedValue(ORDER_KEY, [])
        
        handler = FormSchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals("SomeWidget", field_node.get(ns("widget", self.namespace)))
        self.assertEquals(None, field_node.get(ns("omitted", self.namespace)))
        self.assertEquals("display", field_node.get(ns("mode", self.namespace)))
        self.assertEquals(None, field_node.get(ns("before", self.namespace)))

    def test_write_no_data(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy1")
            
        handler = FormSchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals(None, field_node.get(ns("widget", self.namespace)))
        self.assertEquals(None, field_node.get(ns("omitted", self.namespace)))
        self.assertEquals(None, field_node.get(ns("mode", self.namespace)))
        self.assertEquals(None, field_node.get(ns("before", self.namespace)))

class TestSecuritySchema(unittest.TestCase):
    
    namespace = 'http://namespaces.plone.org/supermodel/security'
    
    def test_read(self):
        field_node = ElementTree.Element('field')
        field_node.set(ns("read-permission", self.namespace), "dummy.Read")
        field_node.set(ns("write-permission", self.namespace), "dummy.Write")
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")
        
        handler = SecuritySchema()
        handler.read(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals({u'dummy': 'dummy.Read'}, IDummy.getTaggedValue(READ_PERMISSIONS_KEY))
        self.assertEquals({u'dummy': 'dummy.Write'}, IDummy.getTaggedValue(WRITE_PERMISSIONS_KEY))
    
    def test_read_no_permissions(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")

        handler = SecuritySchema()
        handler.read(field_node, IDummy, IDummy['dummy'])
        
        self.failIf(READ_PERMISSIONS_KEY in IDummy.getTaggedValueTags())
        self.failIf(WRITE_PERMISSIONS_KEY in IDummy.getTaggedValueTags())
        
    def test_write(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")
        
        IDummy.setTaggedValue(READ_PERMISSIONS_KEY, {u'dummy': 'dummy.Read'})
        IDummy.setTaggedValue(WRITE_PERMISSIONS_KEY, {u'dummy': 'dummy.Write'})
                               
        handler = SecuritySchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals("dummy.Read", field_node.get(ns("read-permission", self.namespace)))
        self.assertEquals("dummy.Write", field_node.get(ns("write-permission", self.namespace)))
    
    def test_write_no_permissions(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")
        
        IDummy.setTaggedValue(READ_PERMISSIONS_KEY, {u'dummy': None})
        
        handler = SecuritySchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals(None, field_node.get(ns("read-permission", self.namespace)))
        self.assertEquals(None, field_node.get(ns("write-permission", self.namespace)))

    def test_write_no_metadata(self):
        field_node = ElementTree.Element('field')
        
        class IDummy(Interface):
            dummy = zope.schema.TextLine(title=u"dummy")
        
        handler = SecuritySchema()
        handler.write(field_node, IDummy, IDummy['dummy'])
        
        self.assertEquals(None, field_node.get(ns("read-permission", self.namespace)))
        self.assertEquals(None, field_node.get(ns("write-permission", self.namespace)))
        
def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFormSchema))
    suite.addTest(unittest.makeSuite(TestSecuritySchema))
    return suite



from z3c.form import field

from plone.z3cform.fieldsets.group import GroupFactory

from plone.autoform.utils import processFieldMoves, processFields

_marker = object()

class AutoFields(object):
    """Mixin class for the WidgetsView and AutoExtensibleForm classes.
    Takes care of actually processing field updates
    """

    schema = None
    additionalSchemata = ()
    
    fields = field.Fields()
    groups = []
    
    ignorePrefix = False
    autoGroups = False
    
    def updateFieldsFromSchemata(self):
 
        # If the form is called from the ++widget++ traversal namespace,
        # we won't have a user yet. In this case, we can't perform permission
        # checks.
        
        have_user = bool(self.request.get('AUTHENTICATED_USER', False))
        
        # Turn fields into an instance variable, since we will be modifying it
        self.fields = field.Fields(self.fields)
        
        # Copy groups to an instance variable and ensure that we have
        # the more mutable factories, rather than 'Group' subclasses

        self.groups = []

        for g in self.groups:
            group_name = getattr(g, '__name__', g.label)
            fieldset_group = GroupFactory(group_name,
                                          field.Fields(g.fields),
                                          g.label,
                                          getattr(g, 'description', None))
            self.groups.append(fieldset_group)
        
        prefixes = {}
        
        # Set up all widgets, modes, omitted fields and fieldsets
        if self.schema is not None:
            processFields(self, self.schema, permissionChecks=have_user)
            for schema in self.additionalSchemata:
                
                # Find the prefix to use for this form and cache for next round
                prefix = self.getPrefix(schema)
                if prefix and prefix in prefixes:
                    prefix = schema.__identifier__
                prefixes[schema] = prefix
                
                # By default, there's no default group, i.e. fields go 
                # straight into the default fieldset
                
                defaultGroup = None
                
                # Create groups from schemata if requested and set default 
                # group

                if self.autoGroups:
                    group_name = schema.__name__
                    
                    # Look for group - note that previous processFields
                    # may have changed the groups list, so we can't easily
                    # store this in a dict.
                    found = False
                    for g in self.groups:
                        if group_name == getattr(g, '__name__', g.label):
                            found = True
                            break
                    
                    if not found:
                        fieldset_group = GroupFactory(group_name,
                                                      field.Fields(),
                                                      group_name,
                                                      schema.__doc__)
                        self.groups.append(fieldset_group)

                    defaultGroup = group_name
                    
                processFields(self, schema, prefix=prefix, defaultGroup=defaultGroup, permissionChecks=have_user)
        
        # Then process relative field movements. The base schema is processed
        # last to allow it to override any movements made in additional 
        # schemata.
        if self.schema is not None:
            for schema in self.additionalSchemata:
                processFieldMoves(self, schema, prefix=prefixes[schema])
            processFieldMoves(self, self.schema)
            
    def getPrefix(self, schema):
        """Get the preferred prefix for the given schema
        """
        if self.ignorePrefix:
            return ''
        return schema.__name__


from zope.component import queryUtility

from zope.security.interfaces import IPermission

from zope.dottedname.resolve import resolve

from z3c.form import field
from z3c.form.util import expandPrefix
from z3c.form.interfaces import IFieldWidget, INPUT_MODE, DISPLAY_MODE

from plone.supermodel.utils import mergedTaggedValueDict
from plone.supermodel.utils import mergedTaggedValueList

from plone.z3cform.fieldsets.group import GroupFactory

from plone.z3cform.fieldsets.utils import move

from plone.supermodel.interfaces import FIELDSETS_KEY

from plone.autoform.interfaces import OMITTED_KEY, WIDGETS_KEY, MODES_KEY, ORDER_KEY
from plone.autoform.interfaces import READ_PERMISSIONS_KEY, WRITE_PERMISSIONS_KEY

from AccessControl import getSecurityManager

_dottedCache = {}

def resolveDottedName(dottedName):
    """Resolve a dotted name to a real object
    """
    global _dottedCache
    if dottedName not in _dottedCache:
        _dottedCache[dottedName] = resolve(dottedName)
    return _dottedCache[dottedName]

def _getDisallowedFields(context, permissions, prefix):
    """Get a list of fields for which the user does not have the requisite
    permission. 'permissions' is a dict with field names as keys and
    permission names as values. The permission names will be looked up
    as IPermission utilities.
    """
    
    security_manager = getSecurityManager()
    
    disallowed = []
    permission_cache = {} # name -> True/False
    
    for fieldName, permission_name in permissions.items():
        if permission_name not in permission_cache:
            permission = queryUtility(IPermission, name=permission_name)
            if permission is None:
                permission_cache[permission_name] = True
            else:
                permission_cache[permission_name] = security_manager.checkPermission(permission.title, context)
        if not permission_cache[permission_name]:
            disallowed.append(_fn(prefix, fieldName))
    
    return disallowed

# Some helper functions

def _fn(prefix, fieldName):
    """Give prefixed fieldname if applicable
    """
    if prefix:
        return expandPrefix(prefix) + fieldName
    else:
        return fieldName

def _bn(field):
    """Give base (non-prefixed) fieldname
    """
    prefix = field.prefix
    fieldName = field.__name__
    if prefix:
        return fieldName[len(prefix) + 1:]
    else:
        return fieldName

def _processWidgets(form, widgets, modes, newFields):
    """Update the fields list with widgets
    """

    for fieldName in newFields:
        field = newFields[fieldName]
        base_name = _bn(field)
        
        widget_name = widgets.get(base_name, None)
        widget_mode = modes.get(base_name, field.mode) or form.mode or INPUT_MODE
        
        widget_factory = None
        if widget_name is not None:
            if isinstance(widget_name, basestring):
                widget_factory = resolveDottedName(widget_name)
            elif IFieldWidget.implementedBy(widget_name):
                widget_factory = widget_name
            
            if widget_factory is not None:
                field.widgetFactory[widget_mode] = widget_factory
        
        if base_name in modes:
            newFields[fieldName].mode = widget_mode

def processFields(form, schema, prefix='', defaultGroup=None, permissionChecks=True):
    """Add the fields from the schema to the from, taking into account
    the hints in the various tagged values as well as fieldsets. If prefix
    is given, the fields will be prefixed with this prefix. If 
    defaultGroup is given (as a Fieldset instance), any field not explicitly
    placed into a particular fieldset, will be added to the given group,
    which must exist already. If permissionChecks is false,
    permission checks are ignored.
    """

    # Get data from tagged values, flattening data from super-interfaces
    
    # Note: The names always refer to a field in the schema, and never contain a prefix.
    
    omitted   = mergedTaggedValueDict(schema, OMITTED_KEY)   # name => e.g. 'true'
    modes     = mergedTaggedValueDict(schema, MODES_KEY)     # name => e.g. 'hidden'
    widgets   = mergedTaggedValueDict(schema, WIDGETS_KEY)   # name => widget/dotted name

    fieldsets = mergedTaggedValueList(schema, FIELDSETS_KEY) # list of IFieldset instances

    # Get either read or write permissions depending on what type of form this is

    permissions = {}
    if permissionChecks:
        if form.mode == DISPLAY_MODE:
            permissions = mergedTaggedValueDict(schema, READ_PERMISSIONS_KEY)  # name => permission name
        elif form.mode == INPUT_MODE:
            permissions = mergedTaggedValueDict(schema, WRITE_PERMISSIONS_KEY) # name => permission name
    
    # Find the fields we should not worry about
    
    groups = {}
    do_not_process = list(form.fields.keys())
    do_not_process.extend(_getDisallowedFields(form.context, permissions, prefix))

    for fieldName, status in omitted.items():
        do_not_process.append(_fn(prefix, fieldName))
    
    for group in form.groups:
        do_not_process.extend(list(group.fields.keys()))
        
        group_name = getattr(group, '__name__', group.label)
        groups[group_name] = group

    # Find all allowed fields so that we have something to select from
    all_fields = field.Fields(schema, prefix=prefix, omitReadOnly=True).omit(*do_not_process)
    
    # Keep track of which fields are in a fieldset, and, by elimination,
    # which ones are not 
    
    fieldset_fields = []
    for fieldset in fieldsets:
        for fieldName in fieldset.fields:
            fieldset_fields.append(_fn(prefix, fieldName))
    
    # Set up the default fields, widget factories and widget modes
    
    newFields = all_fields.omit(*fieldset_fields)
    _processWidgets(form, widgets, modes, newFields)
    
    if not defaultGroup:
        form.fields += newFields
    else:
        groups[defaultGroup].fields += newFields
    
    # Set up fields for fieldsets
    
    for fieldset in fieldsets:
        
        newFields = all_fields.select(*[_fn(prefix, fieldName) 
                                            for fieldName in fieldset.fields
                                                if _fn(prefix, fieldName) in all_fields])
        
        if len(newFields) > 0:        
            _processWidgets(form, widgets, modes, newFields)
        
            if fieldset.__name__ not in groups:
                form.groups.append(GroupFactory(fieldset.__name__,
                                                label=fieldset.label,
                                                description=fieldset.description,
                                                fields=newFields))
            else:
                groups[fieldset.__name__].fields += newFields
    
    
def processFieldMoves(form, schema, prefix=''):
    """Process all field moves stored under ORDER_KEY in the schema tagged
    value. This should be run after all schemata have been processed with
    processFields().
    """
    
    order = mergedTaggedValueList(schema, ORDER_KEY)      # (name, 'before'/'after', other name)
            
    for fieldName, direction, relative_to in order:
        
        # Handle shortcut: leading . means "in this form". May be useful
        # if you want to move a field relative to one in the current
        # schema or (more likely) a base schema of the current schema, without
        # having to repeat the full prefix of this schema.

        if relative_to.startswith('.'):
            relative_to = relative_to[1:]
            if prefix:
                relative_to = expandPrefix(prefix) + relative_to
        
        try:
            if direction == 'before':
                move(form, fieldName, before=relative_to, prefix=prefix)
            elif direction == 'after':
                move(form, fieldName, after=relative_to, prefix=prefix)
        except KeyError:
            # The relative_to field doesn't exist
            pass


from zope.interface import implements

from plone.supermodel.utils import ns
from plone.supermodel.parser import IFieldMetadataHandler

from plone.autoform.interfaces import OMITTED_KEY, WIDGETS_KEY, MODES_KEY, ORDER_KEY
from plone.autoform.interfaces import READ_PERMISSIONS_KEY, WRITE_PERMISSIONS_KEY

from plone.autoform.interfaces import FORM_NAMESPACE, FORM_PREFIX
from plone.autoform.interfaces import SECURITY_NAMESPACE, SECURITY_PREFIX

class FormSchema(object):
    """Support the form: namespace in model definitions.
    """
    implements(IFieldMetadataHandler)
    
    namespace = FORM_NAMESPACE
    prefix = FORM_PREFIX
    
    def _add(self, schema, key, name, value):
        tagged_value = schema.queryTaggedValue(key, {})
        tagged_value[name] = value
        schema.setTaggedValue(key, tagged_value)
    
    def _add_order(self, schema, name, direction, relative_to):
        tagged_value = schema.queryTaggedValue(ORDER_KEY, [])
        tagged_value.append((name, direction, relative_to,))
        schema.setTaggedValue(ORDER_KEY, tagged_value)
    
    def read(self, fieldNode, schema, field):
        name = field.__name__
        
        widget  = fieldNode.get( ns('widget',  self.namespace) )
        mode    = fieldNode.get( ns('mode',    self.namespace) )
        omitted = fieldNode.get( ns('omitted', self.namespace) )
        before  = fieldNode.get( ns('before',  self.namespace) )
        after   = fieldNode.get( ns('after',   self.namespace) )

        if widget:
            self._add(schema, WIDGETS_KEY, name, widget)
        if mode:
            self._add(schema, MODES_KEY, name, mode)
        if omitted:
            self._add(schema, OMITTED_KEY, name, omitted)
        if before:
            self._add_order(schema, name, 'before', before)
        if after:
            self._add_order(schema, name, 'after', after)

    def write(self, fieldNode, schema, field):
        name = field.__name__
        
        widget  = schema.queryTaggedValue(WIDGETS_KEY, {}).get(name, None)
        mode    = schema.queryTaggedValue(MODES_KEY,   {}).get(name, None)
        omitted = schema.queryTaggedValue(OMITTED_KEY, {}).get(name, None)
        order   = [(d,v) for n,d,v in schema.queryTaggedValue(ORDER_KEY,  []) if n == name]
        
        if widget is not None:
            if not isinstance(widget, basestring):
                widget = "%s.%s" % (widget.__module__, widget.__name__)
            fieldNode.set(ns('widget', self.namespace), str(widget))
            
        if mode is not None:
            fieldNode.set(ns('mode', self.namespace), str(mode))
        
        if omitted is not None:
            fieldNode.set(ns('omitted', self.namespace), str(omitted))

        for direction, relative_to in order:
            if direction == 'before':
                fieldNode.set(ns('before',  self.namespace), relative_to)
            elif direction == 'after':
                fieldNode.set(ns('after',  self.namespace), relative_to)

class SecuritySchema(object):
    """Support the security: namespace in model definitions.
    """
    implements(IFieldMetadataHandler)
    
    namespace = SECURITY_NAMESPACE
    prefix = SECURITY_PREFIX
    
    def read(self, fieldNode, schema, field):
        name = field.__name__
        
        read_permission = fieldNode.get(ns('read-permission', self.namespace))
        write_permission = fieldNode.get(ns('write-permission', self.namespace))
        
        read_permissions = schema.queryTaggedValue(READ_PERMISSIONS_KEY, {})
        write_permissions = schema.queryTaggedValue(WRITE_PERMISSIONS_KEY, {})
        
        if read_permission:
            read_permissions[name] = read_permission
            schema.setTaggedValue(READ_PERMISSIONS_KEY, read_permissions)
            
        if write_permission:
            write_permissions[name] = write_permission
            schema.setTaggedValue(WRITE_PERMISSIONS_KEY, write_permissions)

    def write(self, fieldNode, schema, field):
        name = field.__name__
        
        read_permission = schema.queryTaggedValue(READ_PERMISSIONS_KEY, {}).get(name, None)
        write_permission = schema.queryTaggedValue(WRITE_PERMISSIONS_KEY, {}).get(name, None)
        
        if read_permission:
            fieldNode.set(ns('read-permission', self.namespace), read_permission)
        if write_permission:
            fieldNode.set(ns('write-permission', self.namespace), write_permission)


from zope.interface import Interface
from zope.interface.interfaces import IInterface

import zope.schema

from z3c.form.interfaces import IFieldsForm, IDisplayForm, IWidget
from plone.supermodel.interfaces import IFieldset

# Schema interface tagged value keys

OMITTED_KEY   = u"plone.autoform.omitted"
WIDGETS_KEY   = u"plone.autoform.widgets"
MODES_KEY     = u"plone.autoform.modes"
ORDER_KEY     = u"plone.autoform.order"

READ_PERMISSIONS_KEY  = u"plone.autoform.security.read-permissions"
WRITE_PERMISSIONS_KEY = u"plone.autoform.security.write-permissions"

# Supermodel namespace and prefix

FORM_NAMESPACE = 'http://namespaces.plone.org/supermodel/form'
FORM_PREFIX = 'form'

SECURITY_NAMESPACE = 'http://namespaces.plone.org/supermodel/security'
SECURITY_PREFIX = 'security'

class IFormFieldProvider(Interface):
    """Marker interface for schemata that provide form fields.
    """

class IAutoExtensibleForm(Interface):
    """The mixin class plone.autoform.form.AutoExtensibleForm can be
    used to have z3c.form forms that build automatically based on the contents
    of various tagged values, on multiple schema interfaces, which
    should be supplied with the properties defined below. See autoform.txt
    for details.
    """
    
    ignorePrefix = zope.schema.Bool(title=u"Do not set a prefix for additional schemata",
                                    default=False)
    
    schema = zope.schema.Object(title=u"Schema providing form fields",
                                schema=IInterface)
                                
    additionalSchemata = zope.schema.Tuple(title=u"Supplementary schemata providing additional form fields",
                                           value_type=zope.schema.Object(title=u"Schema interface",
                                                                          schema=IInterface),
                                           required=False)

class IWidgetsView(IAutoExtensibleForm, IFieldsForm, IDisplayForm):
    """A display form that supports setting up widgets based on schema
    interfaces.
    """
    
    w = zope.schema.Dict(title=u"Shortcut lookup for all widgets",
                         description=u"Contains all widgets, including "
                                      "those from groups within this form",
                         key_type=zope.schema.ASCIILine(title=u"Widget name, with prefix"),
                         value_type=zope.schema.Object(title=u"Widget", schema=IWidget))

    fieldsets = zope.schema.Dict(title=u"Lookup fieldset (group) by name",
                              key_type=zope.schema.ASCIILine(title=u"Fieldset name"),
                              value_type=zope.schema.Object(title=u"Fieldset", schema=IFieldset))